@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2019 Canonical Ltd.
+ * Copyright (C) 2014-2020 Canonical Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -103,6 +103,7 @@ typedef struct proc_info {
 	pid_t	pid;		/* Process ID */
 	uid_t	uid;		/* User ID */
 	gid_t	gid;		/* GUID */
+	uid_t	euid;		/* EUID */
 	struct timeval start;	/* time when process started */
 	bool	kernel_thread;	/* true if a kernel thread */
 } proc_info_t;
@@ -188,13 +189,17 @@ static unsigned int opt_flags = OPT_CMD_LONG;	/* Default option */
 static int row = 0;				/* tty row number */
 static long int opt_duration = -1;		/* duration, < 0 means run forever */
 
+static char unknown[] = "<unknown>";
+
+static proc_info_t *proc_info_add(const pid_t pid, const struct timeval * const tv);
+
 /* Default void no process info struct */
 static proc_info_t no_info = {
 	.pid = NULL_PID,
 	.uid = NULL_UID,
 	.gid = NULL_GID,
 	.tty = NULL_TTY,
-	.cmdline = "<unknown>",
+	.cmdline = unknown,
 	.kernel_thread = false,
 	.start = { 0, 0 },
 	.next = NULL,
@@ -269,6 +274,20 @@ static const time_scale_t second_scales[] = {
 	{ ' ',  1000,  INT64_MAX },
 };
 
+/*
+ *  proc_comm_dup()
+ *	duplicate a comm filed string, if it fails, return unknown
+ */
+static char *proc_comm_dup(const char *str)
+{
+	char *comm = strdup(str);
+
+	if (!comm)
+		return unknown;
+
+	return comm;
+}
+
 /*
  *  secs_to_str()
  *	report seconds in different units.
@@ -301,21 +320,25 @@ static char *get_username(const uid_t uid)
 	static char buf[12];
 	const size_t hash = uid % MAX_UIDS;
 	uid_name_info_t *uni = uid_name_info[hash];
-	char *name;
+	char *name = NULL;
 
-	/*
-	 *  Try and find it in cache first
-	 */
-	while (uni) {
-		if (uni->uid == uid)
-			return uni->name;
-		uni = uni->next;
+	if (uid != NULL_UID) {
+		/*
+		 *  Try and find it in cache first
+		 */
+		while (uni) {
+			if (uni->uid == uid)
+				return uni->name;
+			uni = uni->next;
+		}
+
+		pwd = getpwuid(uid);
+		if (pwd)
+			name = pwd->pw_name;
 	}
 
-	pwd = getpwuid(uid);
-	if (pwd) {
-		name = pwd->pw_name;
-	} else {
+	/* Unknow UID, use numeric instead */
+	if (!name) {
 		(void)snprintf(buf, sizeof(buf), "%d", uid);
 		name = buf;
 	}
@@ -449,45 +472,125 @@ static void tty_name_info_free(void)
 	}
 }
 
+/*
+ *  proc_name_clean()
+ *	clean up unwanted chars from process name
+ */
+static void proc_name_clean(char *buffer, const int len)
+{
+	char *ptr;
+
+	/*
+	 *  Convert '\r' and '\n' into spaces
+	 */
+	for (ptr = buffer; *ptr && (ptr < buffer + len); ptr++) {
+		if ((*ptr == '\r') || (*ptr =='\n'))
+			*ptr = ' ';
+	}
+	/*
+	 *  OPT_CMD_SHORT option we discard anything after a space
+	 */
+	if (opt_flags & OPT_CMD_SHORT) {
+		for (ptr = buffer; *ptr && (ptr < buffer + len); ptr++) {
+			if (*ptr == ' ') {
+				*ptr = '\0';
+				break;
+			}
+		}
+	}
+}
+
+/*
+ *  proc_comm()
+ *	get process name from comm field
+ */
+static char *proc_comm(const pid_t pid)
+{
+	int fd;
+	ssize_t ret;
+	char buffer[4096];
+
+	(void)snprintf(buffer, sizeof(buffer), "/proc/%d/comm", pid);
+	if ((fd = open(buffer, O_RDONLY)) < 0)
+		return unknown;
+	if ((ret = read(fd, buffer, sizeof(buffer) - 1)) <= 0) {
+		(void)close(fd);
+		return unknown;
+	}
+	(void)close(fd);
+	buffer[ret - 1] = '\0';		/* remove trailing '\n' */
+	proc_name_clean(buffer, ret);
+
+	return proc_comm_dup(buffer);
+}
+
+
 /*
  *  get_extra()
- *	quick and dirty way to get UID and GID from a PID,
- *	note that this does not cater of changes
- *	because of use of an effective ID.
+ *	quick and dirty way to get UID, EUID and GID from a PID
  */
 static void get_extra(const pid_t pid, proc_info_t * const info)
 {
-	ssize_t ret;
 	long dev;
-	int fd;
+	FILE *fp;
 	char path[PATH_MAX];
-	char buffer[4096];
+	char buffer[1024];
 	struct stat buf;
 
 	info->uid = NULL_UID;
 	info->gid = NULL_GID;
+	info->euid = NULL_UID;
 	info->tty = NULL_TTY;
 
 	if (!(opt_flags & OPT_EXTRA))
 		return;
 
-	(void)snprintf(path, sizeof(path), "/proc/%u/stat", pid);
-	fd = open(path, O_RDONLY);
-	if (fd < 0)
+	(void)snprintf(path, sizeof(path), "/proc/%u/status", pid);
+	fp = fopen(path, "r");
+	if (!fp)
 		return;
 
-	if (fstat(fd, &buf) == 0) {
-		info->uid = buf.st_uid;
-		info->gid = buf.st_gid;
+	(void)memset(buffer, 0, sizeof(buffer));
+	while (fgets(buffer, sizeof(buffer) - 1, fp)) {
+		int gid, uid, euid;
+
+		if (!strncmp(buffer, "Uid:", 4)) {
+			if (sscanf(buffer + 4, "%d %d", &uid, &euid) == 2) {
+				info->uid = uid;
+				info->euid = euid;
+			}
+		}
+		if (!strncmp(buffer, "Gid:", 4)) {
+			if (sscanf(buffer + 4, "%d", &gid) == 1) {
+				info->gid = gid;
+			}
+		}
+		if ((info->uid != NULL_UID) &&
+		    (info->euid != NULL_UID) &&
+		    (info->gid != NULL_GID))
+			break;
+	}
+	(void)fclose(fp);
+
+	(void)snprintf(path, sizeof(path), "/proc/%u/stat", pid);
+	fp = fopen(path, "r");
+	if (!fp)
+		return;
+
+	/*
+	 *  Failed to parse /proc/$PID/status? then at least get
+	 *  some info by stat'ing /proc/$PID/stat
+	 */
+	if ((info->uid == NULL_UID) || (info->gid == NULL_GID)) {
+		if (fstat(fileno(fp), &buf) == 0) {
+			info->uid = buf.st_uid;
+			info->gid = buf.st_gid;
+		}
 	}
 
-	ret = read(fd, buffer, sizeof(buffer));
-	(void)close(fd);
-	if (ret < 0)
-		return;
-
-	if (sscanf(buffer, "%*d %*s %*s %*d %*d %*d %ld", &dev) == 1)
+	if (fscanf(fp, "%*d %*s %*s %*d %*d %*d %ld", &dev) == 1)
 		info->tty = (dev_t)dev;
+	(void)fclose(fp);
 }
 
 /*
@@ -715,7 +818,7 @@ static void print_heading(void)
 
 	(void)printf("Time     Event %*.*s %s%sInfo   Duration Process\n",
 		pid_size, pid_size, "PID",
-		(opt_flags & OPT_EXTRA) ? "   UID TTY    " : "",
+		(opt_flags & OPT_EXTRA) ? "    UID    EUID TTY    " : "",
 		(opt_flags & OPT_GLYPH) ? " " : "");
 }
 
@@ -743,6 +846,69 @@ static inline double timeval_to_double(const struct timeval * const tv)
 	return (double)tv->tv_sec + ((double)tv->tv_usec / 1000000.0);
 }
 
+/*
+ *   proc_info_get_timeval()
+ *	get time when process started
+ */
+static void proc_info_get_timeval(const pid_t pid, struct timeval * const tv)
+{
+	int fd;
+	unsigned long long starttime;
+	unsigned long jiffies;
+	char path[PATH_MAX];
+	char buffer[4096];
+	double uptime_secs, secs;
+	struct timeval now;
+	ssize_t n;
+
+	(void)snprintf(path, sizeof(path), "/proc/%d/stat", pid);
+
+	fd = open("/proc/uptime", O_RDONLY);
+	if (fd < 0)
+		return;
+	n = read(fd, buffer, sizeof(buffer));
+	if (n <= 0) {
+		(void)close(fd);
+		return;
+	}
+	(void)close(fd);
+	n = sscanf(buffer, "%lg", &uptime_secs);
+	if (n != 1)
+		return;
+	if (uptime_secs < 0.0)
+		return;
+
+	fd = open(path, O_RDONLY);
+	if (fd < 0)
+		return;
+	n = read(fd, buffer, sizeof(buffer));
+	if (n <= 0) {
+		(void)close(fd);
+		return;
+	}
+	(void)close(fd);
+	n = sscanf(buffer, "%*d %*s %*c %*d %*d %*d %*d %*d %*u %*u %*u %*u "
+		"%*u %*u %*u %*u %*u %*d %*d %*d %*d %llu", &starttime);
+	if (n != 1)
+		return;
+
+	errno = 0;
+	jiffies = sysconf(_SC_CLK_TCK);
+	if (errno)
+		return;
+	secs = uptime_secs - ((double)starttime / (double)jiffies);
+	if (secs < 0.0)
+		return;
+
+	if (gettimeofday(&now, NULL) < 0)
+		return;
+
+	secs = timeval_to_double(&now) - secs;
+	tv->tv_sec = secs;
+	tv->tv_usec = (suseconds_t)secs % 1000000;
+}
+
+
 /*
  *  proc_info_hash()
  * 	hash on PID
@@ -752,6 +918,7 @@ static inline size_t proc_info_hash(const pid_t pid)
 	return pid % MAX_PIDS;
 }
 
+
 /*
  *  proc_name_hash()
  *	Hash a string, from Dan Bernstein comp.lang.c (xor version)
@@ -900,55 +1067,13 @@ static void proc_stats_free(void)
 }
 
 /*
- *  proc_name_clean()
- *	clean up unwanted chars from process name
+ *  free_proc_comm()
+ *	free comm field only if it's not the static unknown string
  */
-static void proc_name_clean(char *buffer, const int len)
+static void free_proc_comm(char *comm)
 {
-	char *ptr;
-
-	/*
- 	 *  Convert '\r' and '\n' into spaces
-	 */
-	for (ptr = buffer; *ptr && (ptr < buffer + len); ptr++) {
-		if ((*ptr == '\r') || (*ptr =='\n')) 
-			*ptr = ' ';
-	}
-	/*
-	 *  OPT_CMD_SHORT option we discard anything after a space
-	 */
-	if (opt_flags & OPT_CMD_SHORT) {
-		for (ptr = buffer; *ptr && (ptr < buffer + len); ptr++) {
-			if (*ptr == ' ') {
-				*ptr = '\0';
-				break;
-			}
-		}
-	}
-}
-
-/*
- *  proc_comm()
- *	get process name from comm field
- */
-static char *proc_comm(const pid_t pid)
-{
-	int fd;
-	ssize_t ret;
-	char buffer[4096];
-
-	(void)snprintf(buffer, sizeof(buffer), "/proc/%d/comm", pid);
-	if ((fd = open(buffer, O_RDONLY)) < 0)
-		return NULL;
-	if ((ret = read(fd, buffer, sizeof(buffer) - 1)) <= 0) {
-		(void)close(fd);
-		return NULL;
-	}
-	(void)close(fd);
-	buffer[ret - 1] = '\0';		/* remove trailing '\n' */
-	proc_name_clean(buffer, ret);
-
-	return strdup(buffer);
+	if (comm != unknown)
+		free(comm);
 }
 
 /*
@@ -961,6 +1086,9 @@ static char *proc_cmdline(const pid_t pid)
 	ssize_t ret;
 	char buffer[4096];
 
+	if (pid == 0)
+		return proc_comm_dup("[swapper]");
+
 	if (opt_flags & OPT_COMM)
 		return proc_comm(pid);
 
@@ -994,9 +1122,9 @@ static char *proc_cmdline(const pid_t pid)
 	proc_name_clean(buffer, ret);
 
 	if (opt_flags & OPT_CMD_DIRNAME_STRIP)
-		return strdup(basename(buffer));
+		return proc_comm_dup(basename(buffer));
 
-	return strdup(buffer);
+	return proc_comm_dup(buffer);
 }
 
 /*
@@ -1007,13 +1135,25 @@ static proc_info_t *proc_info_get(const pid_t pid)
 {
 	const size_t i = proc_info_hash(pid);
 	proc_info_t *info = proc_info[i];
+	struct timeval tv;
 
 	while (info) {
-		if (info->pid == pid)
+		if (info->pid == pid) {
+			if (info->cmdline == unknown)
+				info->cmdline = proc_cmdline(pid);
 			return info;
+		}
 		info = info->next;
 	}
-	return &no_info;
+
+	/* Hrm, not already cached, so get new info */
+	(void)memset(&tv, 0, sizeof(tv));
+	proc_info_get_timeval(pid, &tv);
+	info = proc_info_add(pid, &tv);
+	if (!info)
+		info = &no_info;
+
+	return info;
 }
 
 /*
@@ -1030,7 +1170,8 @@ static void proc_info_free(const pid_t pid)
 			info->pid = NULL_PID;
 			info->uid = NULL_UID;
 			info->gid = NULL_GID;
-			free(info->cmdline);
+			info->euid = NULL_UID;
+			free_proc_comm(info->cmdline);
 			info->cmdline = NULL;
 			return;
 		}
@@ -1051,7 +1192,7 @@ static void proc_info_unload(void)
 
 		while (info) {
 			proc_info_t *next = info->next;
-			free(info->cmdline);
+			free_proc_comm(info->cmdline);
 			free(info);
 			info = next;
 		}
@@ -1070,77 +1211,20 @@ static proc_info_t const *proc_info_update(const pid_t pid)
 	if (info == &no_info)
 		return &no_info;
 	newcmd = proc_cmdline(pid);
-	if (!newcmd)
-		return &no_info;
 
-	free(info->cmdline);
-	info->cmdline = newcmd;
+	/*
+	 *  Don't update if newcmd is unknown, at least
+	 *  we temporarily keep the parent's name or
+	 *  the processes old name
+	 */
+	if (newcmd != unknown) {
+		free_proc_comm(info->cmdline);
+		info->cmdline = newcmd;
+	}
 
 	return info;
 }
 
-/*
- *   proc_info_get_timeval()
- *	get time when process started
- */
-static void proc_info_get_timeval(const pid_t pid, struct timeval * const tv)
-{
-	int fd;
-	unsigned long long starttime;
-	unsigned long jiffies;
-	char path[PATH_MAX];
-	char buffer[4096];
-	double uptime_secs, secs;
-	struct timeval now;
-	ssize_t n;
-
-	(void)snprintf(path, sizeof(path), "/proc/%d/stat", pid);
-
-	fd = open("/proc/uptime", O_RDONLY);
-	if (fd < 0)
-		return;
-	n = read(fd, buffer, sizeof(buffer));
-	if (n <= 0) {
-		(void)close(fd);
-		return;
-	}
-	(void)close(fd);
-	n = sscanf(buffer, "%lg", &uptime_secs);
-	if (n != 1)
-		return;
-	if (uptime_secs < 0.0)
-		return;
-
-	fd = open(path, O_RDONLY);
-	if (fd < 0)
-		return;
-	n = read(fd, buffer, sizeof(buffer));
-	if (n <= 0) {
-		(void)close(fd);
-		return;
-	}
-	(void)close(fd);
-	n = sscanf(buffer, "%*d %*s %*c %*d %*d %*d %*d %*d %*u %*u %*u %*u "
-		"%*u %*u %*u %*u %*u %*d %*d %*d %*d %llu", &starttime);
-	if (n != 1)
-		return;
-
-	errno = 0;
-	jiffies = sysconf(_SC_CLK_TCK);
-	if (errno)
-		return;
-	secs = uptime_secs - ((double)starttime / (double)jiffies);
-	if (secs < 0.0)
-		return;
-
-	if (gettimeofday(&now, NULL) < 0)
-		return;
-
-	secs = timeval_to_double(&now) - secs;
-	tv->tv_sec = secs;
-	tv->tv_usec = (suseconds_t)secs % 1000000;
-}
-
 /*
  *   proc_info_add()
  *	add processes info of a given pid to the hash table
@@ -1152,8 +1236,6 @@ static proc_info_t *proc_info_add(const pid_t pid, const struct timeval * const
 	char *cmdline;
 
 	cmdline = proc_cmdline(pid);
-	if (!cmdline)
-		return NULL;
 
 	/* Re-use any info on the list if it's free */
 	info = proc_info[i];
@@ -1167,7 +1249,7 @@ static proc_info_t *proc_info_add(const pid_t pid, const struct timeval * const
 		info = calloc(1, sizeof(proc_info_t));
 		if (!info) {
 			(void)fprintf(stderr, "Cannot allocate all proc info\n");
-			free(cmdline);
+			free_proc_comm(cmdline);
 			return NULL;
 		}
 		info->next = proc_info[i];
@@ -1255,15 +1337,16 @@ static int proc_info_load(void)
  */
 static char *extra_info(const uid_t uid)
 {
-	static char buf[20];
+	static char buf[28];
 
 	*buf = '\0';
 	if (opt_flags & OPT_EXTRA) {
 		const proc_info_t *info = proc_info_get(uid);
 
 		if (info && info->uid != NULL_UID)
-			(void)snprintf(buf, sizeof(buf), "%6s %-6.6s ",
+			(void)snprintf(buf, sizeof(buf), "%7.7s %7.7s %6.6s ",
 				get_username(info->uid),
+				get_username(info->euid),
 				get_tty(info->tty));
 		else
 			(void)snprintf(buf, sizeof(buf), "%14s", "");
@@ -1631,9 +1714,7 @@ static int monitor(const int sock)
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_COMM)) {
 					pid = proc_ev->event_data.comm.process_pid;
 					info1 = proc_info_get(pid);
-					comm = proc_comm(pid);
-					if (!comm)
-						break;
+					comm = proc_cmdline(pid);
 					row_increment();
 
 					(void)printf("%s comm  %*d %s%s%s       %8s %s%s%s -> %s\n",
@@ -1647,7 +1728,7 @@ static int monitor(const int sock)
 						info1->cmdline,
 						info1->kernel_thread ? "]" : "",
 						comm);
-					free(comm);
+					free_proc_comm(comm);
 					proc_info_update(pid);
 				}
 				break;
